/*------------------------------------------------------------------------*/
// Copyright (C) 2019-2020, Armin Biere, Johannes Kepler University Linz
// Copyright (C) 2020-2025, Alois Zoitl, Johannes Kepler University Linz
// Copyright (C) 2025, .... [ put your name here ] ....
/*------------------------------------------------------------------------*/

#include <stddef.h>
#define PROGRAM_NAME "bogus" // CHANGE THE NAME OF THE PROGRAM!!

/*------------------------------------------------------------------------*/
#include <assert.h>
#include <inttypes.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/resource.h>
#include <sys/time.h>

/*========================================================================*/
// This part is the real 'meat' of the exercise:

static volatile bool go;
static volatile uint64_t global_result;

typedef struct {
    pthread_t thread;
    uint64_t operations;
} WorkerData;

static void waitForGo();
static uint64_t performOneOperation(uint64_t prevResult);

static void *run(void *ptr) {
  WorkerData *myData = ptr;

  waitForGo();

  const uint64_t operations = myData->operations;
  volatile uint64_t *p = &global_result;

  for (uint64_t i = 0; i < operations; i++) {
    *p = performOneOperation(*p);
  }

  return 0;
}

static void waitForGo() {
  while (!go)
    ;
}

static uint64_t performOneOperation(uint64_t prevResult) {
  return prevResult + 1;
}

/*========================================================================*/
// Declaration of helper methods
static void parseCommandLineOptions(int argc, char **argv, size_t *num_workers, uint64_t *total_operations);
static WorkerData *createWorkers(size_t num_workers, uint64_t total_operations);
static void waitForWorkers(WorkerData *workers, size_t num_workers);
static void die(const char *fmt, ...);
static void msg(const char *fmt, ...);
static double wall_clock_time();
static double process_time();

/*------------------------------------------------------------------------*/
int main(int argc, char **argv) {
  size_t num_workers;
  uint64_t total_operations;

  parseCommandLineOptions(argc, argv, &num_workers, &total_operations);

  msg("testing %s version", PROGRAM_NAME);
  msg("initializing %u workers", num_workers);
  msg("executing %" PRIu64 " operations in total", total_operations);

  WorkerData *workers = createWorkers(num_workers, total_operations);

  double w = wall_clock_time();
  double p = process_time();

  // start workers
  go = true;

  waitForWorkers(workers, num_workers);

  free(workers);

  if (global_result == total_operations) {
    msg("SUCCESS: result is %" PRIu64 " as expected", total_operations);
  } else {
    msg("ERROR: result is %" PRIu64 " but expected %" PRIu64, global_result, total_operations);
  }

  w = wall_clock_time() - w;
  p = process_time() - p;

  msg("used %.3f seconds wall-clock and %.3f process time", w, p);
  msg("utilization %.0f%%, %.1f million operations per second", (w ? 100 * p / w : 0),
      w ? total_operations / 1e6 / w : 0);

  return 0;
}

/*------------------------------------------------------------------------*/

static const char *usage = "usage: " PROGRAM_NAME " <workers> <operations>\n"
                           "(operations logarithmically in base 10)\n";

static void parseCommandLineOptions(int argc, char **argv, size_t *num_workers, uint64_t *total_operations) {
  // Parse command line option.
  if (argc != 3) {
    fprintf(stderr, usage, PROGRAM_NAME);
    exit(1);
  }

  int tmp = atoi(argv[1]);
  if (tmp <= 0) {
    die("non-positive number of workers");
  }
  *num_workers = tmp;

  tmp = atoi(argv[2]);
  if (tmp > 20) {
    die("operations exponent too large");
  }
  *total_operations = 1;
  while (tmp--) {
    *total_operations *= 10;
  }
}

static WorkerData *createWorkers(size_t num_workers, uint64_t total_operations) {
  uint64_t rest = total_operations % num_workers;
  uint64_t operations_per_worker = total_operations / num_workers;
  msg("executing %" PRIu64 " operations per worker", operations_per_worker);

  WorkerData *workers = calloc(num_workers, sizeof(*workers));

  for (size_t i = 0; i < num_workers; i++) {
    WorkerData *worker = workers + i;
    worker->operations = operations_per_worker;
    if (rest) {
      worker->operations++;
      rest--;
    }
  }

  for (size_t i = 0; i < num_workers; i++) {
    if (pthread_create(&workers[i].thread, 0, run, workers + i)) {
      die("failed to create worker thread %u", i);
    }
  }
  return workers;
}

static void waitForWorkers(WorkerData *workers, size_t num_workers) {
  for (size_t i = 0; i < num_workers; i++) {
    if (pthread_join(workers[i].thread, 0)) {
      die("failed to join worker thread %u", i);
    }
  }
}

/*------------------------------------------------------------------------*/
// Error and verbose messages.

static void die(const char *fmt, ...) {
  fprintf(stderr, "%s: error: ", PROGRAM_NAME);
  va_list ap;
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  fputc('\n', stderr);
  exit(1);
}

static void msg(const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
  fputc('\n', stdout);
  fflush(stdout);
}

/*------------------------------------------------------------------------*/
// Get current wall-clock and process time in seconds.

static double wall_clock_time() {
  struct timeval tv;
  if (!gettimeofday(&tv, 0)) {
    return 1e-6 * tv.tv_usec + tv.tv_sec;
  }
  return 0.0;
}

static double process_time() {
  struct rusage u;
  if (getrusage(RUSAGE_SELF, &u)) {
    return 0;
  }
  return u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec + u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
}
